(* ----------------------------------------------------------------------- *)
(*									   *)
(* Low-level prettyprinting stream package, based on notes by Greg Nelson. *)
(*									   *)
(* ----------------------------------------------------------------------- *)

functor Pp (structure Wr: WR) : PP =
struct

structure Wr = Wr

open Array List
infix 9 sub

(* ----------------------------------------------------------------------- *)
(*			      Data Structures				   *)
(* ----------------------------------------------------------------------- *)

datatype BreakBehavior =
	NLINDENT of int
      | EXPLICIT of string

datatype Token = 
        STRING of string   (* Never contains \n characters *) 
      | BLANK of int       (* A chunk of empty space *) 
      | SETB               (* Open a new level of breakpoints *)
      | ENDB               (* Close the current level *)
      | BREAK of           (* A possible breakpoint *)
          {always: bool,               (* Always break here? *)
           united: bool,               (* Break if anybody on my level does? *)
           what_to_do: BreakBehavior}  (* What to do if I break *)
      | NEWLINE            (* A low-level linebreak (the high-level routines 
                              use breakpoints to represent linebreaks) *)

type Pp = {(* The output line currently being filled *)
           outbuf: Token array,
           outbuf_nextfree: int ref,
           
           (* Current indentation level *)
           indent: int ref,
           
           (* Current right margin *)
           margin: int ref,
           
           (* Saved state to allow backtracking *)
           saved_inq_curr: int ref,
           saved_outbuf_nextfree: int ref,
           saved_indent: int ref,
           
           (* The low-level stream for formatted output *)
           underlying_wr:Wr.Wr, 
           
           (* The queue of tokens to be processed *)
           inq: Token array,
           inq_nextfree: int ref, 
           inq_curr: int ref, 
           
           (* Where to branch when the input queue gets emptied *)
           empty: unit cont ref,

           (* Where to branch when there's input ready again *)
           nonempty: unit cont ref
          }

val qlen = 1000;
val outbuflen = 1000;
val default_margin = 76;

exception CalledErrorCont
val error_cont : unit cont = 
    callcc (fn k => (callcc (fn ek => throw k ek); raise CalledErrorCont))

exception CoroutineBug
exception PPQueueOverflowed

fun init_pp (wr:Wr.Wr) : Pp = 
    {underlying_wr = wr,
     inq = array(qlen, NEWLINE),
     inq_nextfree = ref 0,
     inq_curr = ref 0,
     saved_inq_curr = ref 0,
     saved_outbuf_nextfree = ref 0,
     saved_indent = ref 0,
     outbuf = array(outbuflen, NEWLINE),
     outbuf_nextfree = ref 0,
     indent = ref 0,
     margin = ref default_margin,
     empty = ref(error_cont), 
     nonempty = ref(error_cont)}


(* ----------------------------------------------------------------------- *)
(*				 Debugging				   *)
(* ----------------------------------------------------------------------- *)

(*(*
   The debugging code has been removed for the moment: executing with it 
   compiled in costs too dearly even if the switch is turned off.  To
   get it back, replace all instances of (*(D with (*(D*) and all instances 
   of D)*) with (*D)*)
*)*)   

val DEBUG = ref false;

(*(D
fun debug ss = if (!DEBUG) 
		   then outputc std_out ((implode ss) ^ "\n")
		   else ()
D)*)

fun brbtostr (NLINDENT(i)) = "NLINDENT(" ^ (makestring i) ^ ")"
  | brbtostr (EXPLICIT(s)) = "EXPLICIT(" ^ s ^ ")"

fun toktostr (BLANK(i)) = "BLANK(" ^ (makestring i) ^ ")"
  | toktostr (NEWLINE) = "NEWLINE"
  | toktostr (SETB) = "SETB"
  | toktostr (ENDB) = "ENDB"
  | toktostr (BREAK{always=always,united=united,what_to_do=what_to_do}) = 
  		       "BREAK{" 
  		     ^ "always=" ^ (if always then "true" else "false") ^ ", "
  		     ^ "united=" ^ (if united then "true" else "false") ^ ", "
  		     ^ "what_to_do=" ^ (brbtostr what_to_do)
  		     ^ "}"
  | toktostr (STRING(s)) = "STRING(" ^ s ^ ")"


(* ----------------------------------------------------------------------- *)
(*			     Queue management				   *)
(* ----------------------------------------------------------------------- *)

fun enqueue (pp:Pp) tok =
    let val {inq=inq, 
             inq_nextfree=inq_nextfree, 
             inq_curr=inq_curr, 
             saved_inq_curr=saved_inq_curr, 
             empty=empty, 
             nonempty=nonempty, 
             ...} = pp
        val curinq_nextfree = !inq_nextfree
        (*(D
	val _ = debug ["enqueue ", toktostr tok, "   ",
				   " inq_nextfree=",makestring (!inq_nextfree),
				   " inq_curr=",makestring (!inq_curr),
				   " saved_inq_curr=",makestring 
				                          (!saved_inq_curr)]
        D)*)
    in
       (*(D debug ["enqueue: ",toktostr tok]; D)*)
       if ((curinq_nextfree+1) mod qlen = (!inq_curr)) 
          orelse ((curinq_nextfree+1) mod qlen = (!saved_inq_curr)) 
          then (outputc std_err ("PP queue overflowed: " 
          	       ^ "\n  !inq_nextfree = " ^ (makestring (!inq_nextfree))
          	       ^ "\n  !inq_curr = " ^ (makestring (!inq_curr))
          	       ^ "\n  !saved_inq_curr = " ^ 
          	              (makestring (!saved_inq_curr))
          	       ^ "\n");
                raise PPQueueOverflowed)
          else ();
       update(inq, curinq_nextfree, tok);
       inq_nextfree := (curinq_nextfree + 1) mod qlen;
       (*(D debug ["enqueue: about to switch"]; D)*)
       callcc (fn k => (empty := k; throw (!nonempty) ()))
       (*(D ; debug ["enqueue: returning"] D)*)
    end

fun requeue (pp:Pp) =
    let val {inq=inq, inq_nextfree=inq_nextfree, 
             inq_curr=inq_curr, empty=empty, nonempty=nonempty, 
             saved_inq_curr=saved_inq_curr, ...} = pp
        (*(D
	val _ = debug ["requeue: "," inq_nextfree=",makestring (!inq_nextfree),
				   " inq_curr=",makestring (!inq_curr),
				   " saved_inq_curr=",
				   makestring (!saved_inq_curr)]
        D)*)
    in 
       inq_curr := ((!inq_curr) - 1) mod qlen
    end

fun dequeue (pp:Pp) =
    let val {inq=inq, 
             inq_nextfree=inq_nextfree, inq_curr=inq_curr, empty=empty, 
             nonempty=nonempty, saved_inq_curr=saved_inq_curr, ...} = pp
        (*(D
	val _ = debug ["dequeue: "," inq_nextfree=",makestring (!inq_nextfree),
				   " inq_curr=",makestring (!inq_curr),
				   " saved_inq_curr=",makestring 
				   (!saved_inq_curr)]
        D)*)
        val _ = (* Front make sure there's something to dequeue *)
      		callcc (fn k => 
			((*D( debug ["dequeue: checking for inq_currut"]; D)*)
			 if (!inq_curr) = (!inq_nextfree) 
			     then ((*D( debug ["dequeue: blocking"]; D)*)
				   nonempty := k; 
				   throw (!empty) ())
			 else ()))
	(*D( val _ = debug ["dequeue: unblocked"] D)*)
	val _ = if (!inq_curr)<0 orelse (!inq_curr)>qlen 
		   then outputc std_err ("PP: About to crash: "
		                         ^(makestring (!inq_curr))^"\n")
		   else ()
	val c = inq sub (!inq_curr)
	val _ = inq_curr := ((!inq_curr) + 1) mod qlen
    in 
       (*D( debug ["dequeue: returning"]; D)*)
       c
    end

(* ----------------------------------------------------------------------- *)
(*				Interaction				   *)
(* ----------------------------------------------------------------------- *)

fun ppsetb pp = enqueue pp SETB

fun ppendb pp = enqueue pp ENDB

fun ppbreak pp b i = 
  enqueue pp (BREAK {always=false, united=b, what_to_do=(NLINDENT i)})

exception NewlineNotAllowedInEXPBREAKString
fun ppexpbreak pp b s = let
  fun check ~1 = ()
    | check i = if substring(s,i,1) = "\n"
                  then raise NewlineNotAllowedInEXPBREAKString
                  else check (i-1)
  in check ((String.length s) - 1);                  
     enqueue pp (BREAK {always=false, united=b, what_to_do=(EXPLICIT s)})
  end

fun forceppbreak pp i = 
  enqueue pp (BREAK {always=true, united=true, what_to_do=(NLINDENT i)})

fun ppwrite pp s = let
     (*D val _ = debug ["write: '", s, "'"] D)*)
     val l = String.length s
     fun dump from to = 
       if from = to then () else enqueue pp (STRING(substring(s,from,to-from)))
     fun loop beg curr =
       if curr = l
         then dump beg curr
       else if substring(s,curr,1) = "\n"
         then (dump beg curr; 
               forceppbreak pp 0; 
               loop (curr+1) (curr+1))
       else loop beg (curr+1)
     in 
         loop 0 0
     end

exception IllegalMargin
fun ppsetmargin (pp as {margin=margin, 
                        outbuf_nextfree=outbuf_nextfree, 
                        ...} : Pp) 
                n =
    if (!outbuf_nextfree) > n orelse n >= outbuflen
       then raise IllegalMargin
       else margin := n


(* ----------------------------------------------------------------------- *)
(*				Processing				   *)
(* ----------------------------------------------------------------------- *)

exception LineTooLong
exception HowdThatGetInHere

fun raw_printline (pp as {underlying_wr=underlying_wr, outbuf=outbuf, 
                          outbuf_nextfree=outbuf_nextfree, ...}:Pp) =
      let val i = ref 0
          fun writeblank 0 = ()
            | writeblank count = (Wr.write_wr underlying_wr " "; 
                                  writeblank (count-1))
      in 
         while ((!i)<(!outbuf_nextfree)) do
           (case outbuf sub (!i) of
                NEWLINE => Wr.write_wr underlying_wr "\n"
              | STRING(s) => Wr.write_wr underlying_wr s
              | BLANK(count) => writeblank count
              | _ => raise HowdThatGetInHere;
            i := (!i)+1);
         Wr.flush underlying_wr
      end 

fun write_tok (pp as {outbuf=outbuf, outbuf_nextfree=outbuf_nextfree, 
		      indent=indent, margin=margin, ...}:Pp) 
	      c raiseok =
    let in
	update (outbuf,!outbuf_nextfree,c);
	if (!outbuf_nextfree) < outbuflen 
	   then outbuf_nextfree := (!outbuf_nextfree)+1
	   else ();
	case c of
	    NEWLINE => (raw_printline pp;
			indent := 0;
			outbuf_nextfree := 0)
	  | STRING(s)  => (indent := (!indent)+(String.length s);
			 if (!indent) > (!margin) andalso raiseok
			    then raise LineTooLong 
		            else ())
	  | BLANK(i) => (indent := (!indent) + i;
			 if (!indent) > (!margin) andalso raiseok
			    then raise LineTooLong 
		            else ())
          | _ => raise HowdThatGetInHere
    end
    
fun do_break pp indent (NLINDENT(n)) = let
    (*D( val _ = debug ["do_break: NLINDENT(", makestring n, ")"] D)*)
    in
      write_tok pp NEWLINE false;
      write_tok pp (BLANK(if n+indent > 0 then n+indent else 0)) true
    end
  | do_break pp indent (EXPLICIT(s)) = let
    (*D( val _ = debug ["do_break: EXPLICIT(", s, ")"] D)*)
    in write_tok pp (STRING(s)) false
    end

(* Set everything up to the first unmatched ENDB on the same line, ignoring
   breakpoints. If it won't fit or contains an internal forced break, 
   raise LineTooLong.  Consume the closing ENDB. *)
fun P_one pp = let 
  fun loop() = 
    case dequeue pp of
        SETB => (P_one pp; loop())
      | ENDB => ()
      | BREAK({always=false, ...}) => loop()
      | BREAK({always=true, ...}) => raise LineTooLong
      | NEWLINE => raise HowdThatGetInHere
      | c => (write_tok pp c true; loop())
  in loop() end

(* Set everything up to the first top-level BREAK or unmatched ENDB on the 
   same line, raising LineTooLong if this is not possible.  Leave the 
   input pointer before the BREAK or ENDB. *)
fun P_one_b pp = let 
  fun loop() = 
    case dequeue pp of
        SETB => (P_one pp; loop())
      | ENDB => ()
      | BREAK(_) => ()
      | NEWLINE => raise HowdThatGetInHere
      | c => (write_tok pp c true; loop())
  in 
    loop();
    requeue pp
  end

fun save_state (pp as {saved_inq_curr=saved_inq_curr, 
	               saved_outbuf_nextfree=saved_outbuf_nextfree, 
	               saved_indent=saved_indent,
	               inq_curr=inq_curr, 
                       outbuf_nextfree=outbuf_nextfree, 
                       indent=indent, 
	               ...}:Pp) = 
  (saved_inq_curr := (!inq_curr); 
   saved_outbuf_nextfree := (!outbuf_nextfree); 
   saved_indent := (!indent))

fun restore_state (pp as {saved_inq_curr=saved_inq_curr, 
	                  saved_outbuf_nextfree=saved_outbuf_nextfree, 
	                  saved_indent=saved_indent,
	                  inq_curr=inq_curr, 
                          outbuf_nextfree=outbuf_nextfree, 
                          indent=indent, 
	                  ...}:Pp) = 
  (inq_curr := (!saved_inq_curr); 
   outbuf_nextfree := (!saved_outbuf_nextfree); 
   indent := (!saved_indent))

(* Reset the saved inq pointer so that the enqueue routine will know that 
   we will not be needing to backtrack and it's safe to keep adding
   tokens to the buffer until the free pointer runs into the current
   input pointer.  (Otherwise it has to stop when it reaches the
   saved input pointer.) *)
fun no_going_back (pp as {saved_inq_curr=saved_inq_curr, ...}:Pp) = 
  saved_inq_curr := ~1 

(* Set a group on multiple lines, breaking at all united breakpoints and 
   Returns when an unmatched ENDB is reached.  The final ENDB is consumed. 
   The LineTooLong exception cannot be raised here: if we're already breaking
   all the groups from here out, then there's no point paying attention
   to whether we're overflowing the margin; just do it. *)
fun P_many (pp as {indent=indent, ...} :Pp) = let 
  val current_indent = !indent
  fun loop() = 
    case dequeue pp of
        SETB => (P_one_or_many pp; loop())
      | BREAK({united=true,what_to_do=what_to_do, ...}) => 
            (do_break pp current_indent what_to_do;
             no_going_back pp; 
             loop())
      | BREAK({united=false,what_to_do=what_to_do, ...}) => 
            (save_state pp;
             (P_one_b pp)
                handle LineTooLong 
                   => (restore_state pp;
                       do_break pp current_indent what_to_do);
             no_going_back pp; 
             loop())
      | ENDB => ()
      | NEWLINE => raise HowdThatGetInHere
      | c => (* Just do it: *) 
             (write_tok pp c false; loop())
  in loop() end

(* Try to set a group (up to an unmatched ENDB) on the same line,
   using P_one.  If this causes a LineTooLong exception, then backtrack 
   to the starting position and set the group, using P_many, on multiple 
   lines. *)
and P_one_or_many pp = 
 (save_state pp;
  (P_one pp)
     handle LineTooLong 
       => (restore_state pp;
           P_many pp);
  no_going_back pp)

exception EndbWithNoMatchingSetb
fun top_level pp = 
  (P_many pp;
   raise EndbWithNoMatchingSetb)


(* ----------------------------------------------------------------------- *)
(*				  Creation				   *)
(* ----------------------------------------------------------------------- *)

fun pp_from_wr wr =
    let (*D( val _ = debug ["new"]; D)*)
        val (pp as {empty=empty, ...}:Pp) = init_pp wr
    in 
       callcc (fn k => (empty := k; top_level pp));
       pp
    end

fun wr_from_pp (pp as {underlying_wr=underlying_wr, ...} : Pp)
    = Wr.to_fn 
         (fn s => ppwrite pp s) 
    	 (fn () => (forceppbreak pp 0; 
    	            Wr.flush underlying_wr; 
    	            Wr.close underlying_wr))
    	 (fn () => (forceppbreak pp 0; 
    	            Wr.flush underlying_wr))


end (* Functor Pp *)

