functor Simplify(
	    structure Globals: GLOBALS
	    structure Typ: TYP
	    structure Leq: LEQ
	    structure TEnv: TENV
	    structure TypHelper: TYPHELPER
	    sharing Typ.Globals = Globals
	        and Leq.Typ = Typ
	        and TypHelper.TEnv = TEnv
	        and TEnv.Typ = Typ
	        and Leq.TEnv = TEnv
	        and Typ.Id = Globals.Id
	    ) : SIMPLIFY = struct

structure Globals = Globals
structure Typ = Typ
structure Leq = Leq
structure TEnv = TEnv
open Globals
open TypHelper
open TEnv
open Typ

val DEBUG = createflag ["simplify", "Simplify"] false

fun flatten [] = []
  | flatten (hd::tl) = 
      case explode(hd) of
            EXPTVAR _ => hd :: (flatten tl)
          | EXPARROW _ => hd :: (flatten tl)
          | EXPALL _ => hd :: (flatten tl)
          | EXPSOME _ => hd :: (flatten tl)
          | EXPMEET ts => (flatten ts) @ (flatten tl)
          | EXPOBJ _ => hd :: (flatten tl)
          | EXPMINUS _ => hd :: (flatten tl)

fun pack te ts = let
  fun s befr [] = befr
    | s befr (t::aftr) = if Leq.leq te (meet (befr@aftr)) t
    			   then s befr aftr
    			   else s (befr @ [t]) aftr
  in s [] ts
  end

fun packmeet te ts = let 
  val flat = flatten ts
  val packed = pack te (pack te flat)
  in case packed of
       [t] => t
     | _   => meet packed
  end

fun simplify te t =
  case explode(t ) of
     EXPTVAR a => t
   | EXPARROW (t1,t2) => let
       val t1s = simplify te t1
       val t2s = simplify te t2
       in case explode(t2s) of
             EXPTVAR _ => arrow t1s t2s
           | EXPARROW _ => arrow t1s t2s
           | EXPALL _ => arrow t1s t2s
           | EXPSOME _ => arrow t1s t2s
           | EXPMEET (t2ss) => packmeet te 
           			  (map (fn t2si => arrow t1s t2si) t2ss)
           | EXPOBJ _ => arrow t1s t2s
           | EXPMINUS _ => arrow t1s t2s
       end
   | EXPALL (n,t1,t2) => let
       val t1s = simplify te t1
       val t2s = simplify (push_bound te n t1) t2
       in case explode(t2s) of
             EXPTVAR _ => all n t1s t2s
           | EXPARROW _ => all n t1s t2s
           | EXPALL _ => all n t1s t2s
           | EXPSOME _ => all n t1s t2s
           | EXPMEET (t2ss) => packmeet te (map (fn t2si => all n t1s t2si) t2ss)
           | EXPOBJ _ => all n t1s t2s
           | EXPMINUS _ => all n t1s t2s
       end
   | EXPSOME (n,t1,t2) => let
       val t1s = simplify te t1
       val t2s = simplify (push_bound te n t1) t2
       in some n t1s t2s
       end
   | EXPMEET ts => packmeet te (map (fn t => simplify te t) ts)
   | EXPOBJ (n,t1) => let
       val t1s = simplify te t1
       in case explode(t1s) of
             EXPTVAR _ => obj n t1s
           | EXPARROW _ => obj n t1s
           | EXPALL _ => obj n t1s
           | EXPSOME _ => obj n t1s
           | EXPMEET (t1ss) => packmeet te (map (fn t1si => obj n t1si) t1ss)
           | EXPOBJ _ => obj n t1s
           | EXPMINUS _ => obj n t1s
       end
   | EXPMINUS (t1,ids) => let
       val t1s = simplify te t1
       in case explode(t1s) of
             EXPTVAR _ => minus t1s ids
           | EXPARROW _ => t1s
           | EXPALL _ => t1s
           | EXPSOME _ => t1s
           | EXPMEET (t1ss) => packmeet te 
	                 (* $$$ could be improved by finer analysis *)
	                 (map (fn t1si => simplify te (minus t1si ids)) t1ss)
           | EXPOBJ (l,t1s1) => 
	     if memq Id.== ids l
	       then meet []
	       else t1s
           | EXPMINUS (t1s1,ids1) => minus t1s1 (ids @ ids1)
       end

fun unabbrev te t =
  case explode(t ) of
     EXPTVAR a => lookup_and_relocate te a
       (fn _ => t)
       (fn _ => raise WrongKindOfId(te,a,"tvar or tabbrev"))
       (fn ab => unabbrev te ab)
   | EXPARROW (t1,t2) => arrow (unabbrev te t1) (unabbrev te t2)
   | EXPALL (n,t1,t2) => all n (unabbrev te t1) (unabbrev (push_bound te n t1) t2)
   | EXPSOME (n,t1,t2) => some n (unabbrev te t1)(unabbrev (push_bound te n t1) t2)
   | EXPMEET ts => meet (map (fn t => unabbrev te t) ts)
   | EXPOBJ (n,t1) => obj n (unabbrev te t1)
   | EXPMINUS (t1,ids) => minus (unabbrev te t1) ids

fun undist_meet te ts = let
  fun add_arrow t1 t2 [] = [(t1,[t2])]
    | add_arrow t1 t2 ((fst as (t1',t2s))::rest) = 
      if Leq.eqv te t1 t1'
         then ((t1',t2::t2s)::rest)
         else (fst::(add_arrow t1 t2 rest))
  fun add_all n t1 t2 [] = [(n,t1,[t2])]
    | add_all n t1 t2 ((fst as (n',t1',t2s))::rest) = 
      if Leq.eqv te t1 t1'
         then ((n,t1',t2::t2s)::rest)
         else (fst::(add_all n t1 t2 rest))
  fun add_obj n t2 [] = [(n,[t2])]
    | add_obj n t2 ((fst as (n',t2s))::rest) = 
      if Id.== n n'
         then ((n',t2::t2s)::rest)
         else (fst::(add_obj n t2 rest))
  fun add_minus t1 ids minuses = 
      (* I think, for this case, not smashing is what's wanted: *)
      (ids,[t1])::minuses 
  fun collect [] = ([], [], [], [], [])
    | collect (t::ts) = let
        val (others, arrows, alls, objs, minuses) = collect ts
        in case explode(t ) of
	      EXPTVAR a => 
	     	(t::others, arrows, alls, objs, minuses)
	    | EXPARROW (t1,t2) => 
	     	(others, add_arrow t1 t2 arrows, alls, objs, minuses)
	    | EXPALL (n,t1,t2) => 
	     	(others, arrows, add_all n t1 t2 alls, objs, minuses)
	    | EXPSOME (n,t1,t2) => 
	     	(t::others, arrows, alls, objs, minuses)
	    | EXPMEET ts => 
	     	raise CantHappen
	    | EXPOBJ (n,t1) => 
	     	(others, arrows, alls, add_obj n t1 objs, minuses)
	    | EXPMINUS (t1,ids) => 
	     	(others, arrows, alls, objs, add_minus t1 ids minuses)
        end
  val (others, arrows, alls, objs, minuses) = collect ts
  val res_others = others
  val res_arrows = map
  		     (fn (t1,t2s) => arrow t1 (undist_meet te t2s))
  		     arrows
  val res_alls = map
  		     (fn (n,t1,t2s) => 
  		     	   all n t1 (undist_meet (push_bound te n t1) t2s))
  		     alls
  val res_objs = map
  		     (fn (n,t2s) => obj n (undist_meet te t2s))
  	             objs
  val res_minuses = map
  		     (fn (ids,t2s) => minus (undist_meet te t2s) ids)
  	             minuses
  val results = (res_others @ res_arrows @ res_alls @ res_objs @ res_minuses)
  in 
     case results of
         [t] => t
       | ts  => meet ts
  end

fun undist te t =
  case explode(t ) of
        EXPTVAR a => t
      | EXPARROW (t1,t2) => arrow (undist te t1) (undist te t2)
      | EXPALL (n,t1,t2) => all n (undist te t1) (undist (push_bound te n t1) t2)
      | EXPSOME (n,t1,t2) => some n (undist te t1) (undist (push_bound te n t1) t2)
      | EXPMEET ts => undist_meet te (map (fn t => undist te t) ts)
      | EXPOBJ (n,t1) => obj n (undist te t1)
      | EXPMINUS (t1,ids) => minus (undist te t1) ids

fun squash te t = undist te (simplify te (unabbrev te t))

end

